from __future__ import annotations

from django.http import HttpRequest

from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import PermissionDenied

from allauth.idp.oidc.internal.oauthlib.server import get_server
from allauth.idp.oidc.internal.oauthlib.utils import (
    extract_params,
    get_validator_context,
)
from allauth.idp.oidc.internal.resources import is_resources_subset


class TokenAuthentication(BaseAuthentication):
    """
    Use the OIDC access token to authenticate the request.
    """

    def authenticate(self, request: HttpRequest):
        server = get_server()
        orequest = extract_params(request)
        valid, orequest = server.verify_request(*orequest, scopes=[])
        if not valid:
            return None
        access_token = get_validator_context().access_token
        if access_token:
            resources = access_token.get_resources()
            if not is_resources_subset(
                [request.build_absolute_uri(request.path)], resources
            ):
                raise PermissionDenied("Invalid target resource.")

        return orequest.user, access_token



from __future__ import annotations

from django.core.management.base import BaseCommand

from allauth.idp.oidc.models import Token


class Command(BaseCommand):
    help = "Deletes expired OpenID Connect tokens."

    def handle(self, *args, **options) -> None:
        count, _ = Token.objects.expired().delete()
        self.stdout.write(f"{count} expired token(s) deleted.")



from __future__ import annotations

import uuid
from datetime import timedelta
from typing import Any

from django.utils import timezone

from oauthlib.common import Request
from oauthlib.openid import RequestValidator

from allauth.core import context
from allauth.core.internal import httpkit, jwkkit, ratelimit
from allauth.core.internal.deferred import jwt
from allauth.idp.oidc import app_settings
from allauth.idp.oidc.adapter import get_adapter
from allauth.idp.oidc.internal.clientkit import (
    is_origin_allowed,
    is_redirect_uri_allowed,
    lookup_client,
)
from allauth.idp.oidc.internal.oauthlib import authorization_codes
from allauth.idp.oidc.internal.oauthlib.utils import (
    ValidatorContext,
    get_validator_context,
)
from allauth.idp.oidc.internal.resources import InvalidTargetError, is_resources_subset
from allauth.idp.oidc.internal.tokens import decode_jwt_token, determine_token_type
from allauth.idp.oidc.models import Client, Token


class OAuthLibRequestValidator(RequestValidator):
    def validate_client_id(self, client_id: Any, request: Request) -> bool:
        if not isinstance(client_id, str):
            return False
        client = self._lookup_client(request, client_id)
        if not client:
            return False
        self._use_client(request, client)
        return True

    def validate_redirect_uri(
        self, client_id, redirect_uri, request, *args, **kwargs
    ) -> bool:
        return is_redirect_uri_allowed(
            redirect_uri,
            request.client.get_redirect_uris(),
            request.client.allow_uri_wildcards,
        )

    def validate_response_type(
        self, client_id, response_type, client, request: Request, *args, **kwargs
    ) -> bool:
        return response_type in request.client.get_response_types()

    def validate_scopes(
        self, client_id, scopes, client, request: Request, *args, **kwargs
    ) -> bool:
        return set(scopes).issubset(request.client.get_scopes())

    def get_default_scopes(
        self, client_id, request: Request, *args, **kwargs
    ) -> list[str]:
        return request.client.get_default_scopes()

    def save_authorization_code(
        self, client_id, code, request: Request, *args, **kwargs
    ) -> None:
        # WORKAROUND: docstring says:
        # > To support OIDC, you MUST associate the code with:
        # > - nonce, if present (``code["nonce"]``)
        # Yet, nonce is not there, it is in request.nonce.
        nonce = getattr(request, "nonce", None)
        if nonce:
            code = dict(**code, nonce=nonce)
        # (end WORKAROUND)
        authorization_codes.create(request.client, code, request)

    def authenticate_client_id(
        self, client_id, request: Request, *args, **kwargs
    ) -> bool:
        """Ensure client_id belong to a non-confidential client."""
        if (
            Client.AuthenticationMethod.NONE
            not in get_validator_context().supported_auth_methods
        ):
            return False
        client = self._lookup_client(request, client_id)
        if not client or client.type != Client.Type.PUBLIC:
            return False
        self._use_client(request, client)
        return True

    def authenticate_client(self, request: Request, *args, **kwargs) -> bool:
        ctx = get_validator_context()
        client_id, client_secret = httpkit.extract_basic_auth(request.headers)
        if (
            Client.AuthenticationMethod.CLIENT_SECRET_POST in ctx.supported_auth_methods
            and not client_id
            and not client_secret
        ):
            client_id = getattr(request, "client_id", None)
            client_secret = getattr(request, "client_secret", None)
        if not isinstance(client_id, str):
            return False
        if not client_secret and request.grant_type == Client.GrantType.DEVICE_CODE:
            return self.authenticate_client_id(client_id, request)
        if not client_secret or not isinstance(client_secret, str):
            return False
        client = self._lookup_client(request, client_id)
        if not client:
            return False
        if not client.check_secret(client_secret):
            return False
        self._use_client(request, client)
        return True

    def validate_grant_type(
        self, client_id, grant_type, client, request: Request, *args, **kwargs
    ) -> bool:
        return grant_type in client.get_grant_types()

    def validate_code(
        self, client_id, code, client, request: Request, *args, **kwargs
    ) -> bool:
        return authorization_codes.validate(client_id, code, request)

    def introspect_token(
        self, token, token_type_hint, request: Request, *args, **kwargs
    ) -> dict | None:
        """
        https://datatracker.ietf.org/doc/html/rfc7662#section-2.1
        > The protected resource calls the introspection endpoint using an HTTP
        > POST [RFC7231] request with parameters sent as
        > "application/x-www-form-urlencoded" data as defined in
        > [W3C.REC-html5-20141028].  The protected resource sends a parameter
        > representing the token along with optional parameters representing
        > additional context that is known by the protected resource to aid the
        > authorization server in its response.

        > token
        >     REQUIRED.  The string value of the token.  For access tokens, this
        >     is the "access_token" value returned from the token endpoint
        >     defined in OAuth 2.0 [RFC6749], Section 5.1.  For refresh tokens,
        >     this is the "refresh_token" value returned from the token endpoint
        >     as defined in OAuth 2.0 [RFC6749], Section 5.1.  Other token types
        >     are outside the scope of this specification.

        > token_type_hint
        >     OPTIONAL.  A hint about the type of the token submitted for
        >     introspection.  The protected resource MAY pass this parameter to
        >     help the authorization server optimize the token lookup.  If the
        >     server is unable to locate the token using the given hint, it MUST
        >     extend its search across all of its supported token types.  An
        >     authorization server MAY ignore this parameter, particularly if it
        >     is able to detect the token type automatically.  Values for this
        >     field are defined in the "OAuth Token Type Hints" registry defined
        >     in OAuth Token Revocation [RFC7009].
        """
        ratelimit.consume(
            request=context.request,
            config=app_settings.RATE_LIMITS,
            action="introspect_client",
            key=request.client.id,
            raise_exception=True,
        )
        types, fallback_types = determine_token_type(token, token_type_hint)
        if types is None:
            return None

        instance = Token.objects.lookup_by_value(types, token)
        if not instance and fallback_types:
            instance = Token.objects.lookup_by_value(fallback_types, token)
        if not instance:
            return None

        adapter = get_adapter()
        if instance.user and not instance.user.is_active:
            return None

        if (
            not app_settings.INTROSPECTION_CROSS_CLIENT_ALLOWED
            and instance.client_id != request.client.pk
        ):
            return None

        if not adapter.is_introspection_allowed(
            token=instance,
            caller_client=request.client,
        ):
            """
            https://datatracker.ietf.org/doc/html/rfc7662#section-2.3
            > Note that a properly formed and authorized query for an inactive or
            > otherwise invalid token (or a token the protected resource is not
            > allowed to know about) is not considered an error response by this
            > specification.  In these cases, the authorization server MUST instead
            > respond with an introspection response with the "active" field set to
            > "false" as described in Section 2.2.
            """
            return None

        claims = {
            "iss": adapter.get_issuer(),
            "iat": int(instance.created_at.timestamp()),
            "token_type": (
                "refresh_token"
                if instance.type == Token.Type.REFRESH_TOKEN
                else "access_token"
            ),
            "scope": " ".join(instance.get_scopes()),
        }
        if aud := instance.get_resources():
            claims["aud"] = aud
        if instance.expires_at:
            claims["exp"] = int(instance.expires_at.timestamp())
        if instance.user and instance.client:
            claims["sub"] = adapter.get_user_sub(instance.client, instance.user)
        if instance.client:
            claims["client_id"] = instance.client.id
        adapter.populate_introspection_response(response=claims, token=instance)
        return claims

    def confirm_redirect_uri(
        self, client_id, code, redirect_uri, client, request: Request, *args, **kwargs
    ) -> bool:
        authorization_code = self._lookup_authorization_code(request, client_id, code)
        if not authorization_code:
            return False
        return redirect_uri == authorization_code["redirect_uri"]

    def save_bearer_token(self, token: dict, request: Request, *args, **kwargs) -> None:
        """
        https://datatracker.ietf.org/doc/html/rfc6749#section-6
        > The authorization server MAY issue a new refresh token, in which case
        > the client MUST discard the old refresh token and replace it with the
        > new refresh token.  The authorization server MAY revoke the old
        > refresh token after issuing a new refresh token to the client.  If a
        > new refresh token is issued, the refresh token scope MUST be
        > identical to that of the refresh token included by the client in the
        > request.

        https://datatracker.ietf.org/doc/html/rfc6749#section-1.5
        > Refresh tokens are issued to the client by the authorization server and
        > are used to obtain a new access token when the current access token becomes
        > invalid or expires, or to obtain additional access tokens with identical or
        > narrower scope
        """
        ctx = get_validator_context()
        refresh_token = token.get("refresh_token")
        tokens = []
        if refresh_token:
            rt = self._prep_refresh_token(ctx, refresh_token, request)
            if not rt.pk:
                tokens.append(rt)
            if rt.expires_at is not None:
                # Expose the refresh token lifetime to the client. Not part of
                # RFC 6749, but ``refresh_expires_in`` is the de-facto convention
                # (e.g. Keycloak). Omitted when refresh tokens do not expire.
                token["refresh_expires_in"] = max(
                    0, int((rt.expires_at - timezone.now()).total_seconds())
                )
        access_token = self._prep_access_token(ctx, token, request)
        tokens.append(access_token)
        for t in tokens:
            t.set_scopes(request.scopes)
            if ctx.email:
                t.set_scope_email(ctx.email)
        Token.objects.bulk_create(tokens)

    def _prep_access_token(
        self, ctx: ValidatorContext, token: dict, request: Request
    ) -> Token:
        adapter = get_adapter()
        access_token = Token(
            client=request.client,
            user=request.user,
            type=Token.Type.ACCESS_TOKEN,
            hash=adapter.hash_token(token["access_token"]),
            expires_at=timezone.now() + timedelta(seconds=token["expires_in"]),
        )
        if resources := ctx.requested_resources:
            if not is_resources_subset(resources, ctx.granted_resources):
                raise InvalidTargetError
        else:
            resources = ctx.granted_resources
        if resources:
            access_token.set_resources(resources)
        return access_token

    def _prep_refresh_token(
        self, ctx: ValidatorContext, refresh_token: str, request: Request
    ) -> Token:
        adapter = get_adapter()
        refresh_token_hash = adapter.hash_token(refresh_token)
        rt: Token | None = ctx.refresh_token
        if rt:
            ctx.granted_resources = rt.get_resources()
            if not ctx.email and "email" in request.scopes:
                ctx.email = rt.get_scope_email()
        if (
            rt
            and not app_settings.ROTATE_REFRESH_TOKEN
            and refresh_token_hash == rt.hash
        ):
            # We reuse our token.
            return rt

        resources: list[str] | None
        if rt:
            resources = rt.get_resources()
            # If we have an existing refresh token, drop it, because of:
            assert (
                app_settings.ROTATE_REFRESH_TOKEN or refresh_token_hash != rt.hash
            )  # nosec[assert_used]
            rt.delete()
            rt = None
        else:
            resources = ctx.granted_resources
        refresh_expires_in = app_settings.REFRESH_TOKEN_EXPIRES_IN
        new_rt = Token(
            client=request.client,
            user=request.user,
            type=Token.Type.REFRESH_TOKEN,
            hash=refresh_token_hash,
            expires_at=(
                timezone.now() + timedelta(seconds=refresh_expires_in)
                if refresh_expires_in
                else None
            ),
        )
        if resources:
            new_rt.set_resources(resources)
        rt = new_rt
        return rt

    def invalidate_authorization_code(
        self, client_id, code, request: Request, *args, **kwargs
    ) -> None:
        authorization_codes.invalidate(client_id, code)

    def validate_user_match(
        self, id_token_hint, scopes, claims, request: Request
    ) -> bool:
        if not context.request.user:
            return False
        sub = None
        if id_token_hint:
            payload = decode_jwt_token(
                id_token_hint,
                client_id=request.client.id,
                verify_exp=True,
                verify_iss=True,
            )
            if payload is None:
                return False
            sub = payload.get("sub")
            session_sub = get_adapter().get_user_sub(
                request.client, context.request.user
            )
            if sub != session_sub:
                return False
        if claims:
            sub = claims.get("sub")
            session_sub = get_adapter().get_user_sub(
                request.client, context.request.user
            )
            if sub != session_sub:
                return False
        return True

    def get_authorization_code_scopes(
        self, client_id, code, redirect_uri, request
    ) -> list[str]:
        authorization_code = self._lookup_authorization_code(request, client_id, code)
        if not authorization_code:
            return []
        return authorization_code["scopes"]

    def get_authorization_code_nonce(
        self, client_id, code, redirect_uri, request
    ) -> str | None:
        authorization_code = self._lookup_authorization_code(request, client_id, code)
        if authorization_code is None:
            return None
        return authorization_code["code"].get("nonce")

    def get_code_challenge(self, code, request: Request) -> str | None:
        ret = None
        authorization_code = self._lookup_authorization_code(
            request, request.client_id, code
        )
        if authorization_code is None:
            return None
        if pkce := authorization_code.get("pkce"):
            ret = pkce["code_challenge"]
        return ret

    def get_code_challenge_method(self, code, request: Request) -> str | None:
        ret = None
        authorization_code = self._lookup_authorization_code(
            request, request.client_id, code
        )
        if authorization_code is None:
            return None
        if pkce := authorization_code.get("pkce"):
            ret = pkce["code_challenge_method"]
        return ret

    def is_pkce_required(self, client_id, request: Request) -> bool:
        client = self._lookup_client(request, client_id)
        return bool(client and client.type == Client.Type.PUBLIC)

    def finalize_id_token(
        self, id_token: dict, token: dict, token_handler, request
    ) -> str:
        """
        https://openid.net/specs/openid-connect-core-1_0.html#StandardClaims
        """
        adapter = get_adapter()
        id_token["iss"] = adapter.get_issuer()
        id_token["exp"] = id_token["iat"] + app_settings.ID_TOKEN_EXPIRES_IN
        id_token["jti"] = uuid.uuid4().hex
        ctx = get_validator_context()
        email = ctx.email
        id_token.update(
            adapter.get_claims(
                "id_token", request.user, request.client, request.scopes, email=email
            )
        )
        adapter.populate_id_token(id_token, request.client, request.scopes)
        jwk_dict, private_key = jwkkit.load_jwk_from_pem(adapter.get_signing_key().pem)
        return jwt.encode(
            id_token, private_key, algorithm="RS256", headers={"kid": jwk_dict["kid"]}
        )

    def validate_bearer_token(self, token, scopes, request: Request) -> bool:
        if not token:
            return False
        if context.request.GET.get("access_token") == token:
            # Supporting tokens in query params is considered bad practice, yet,
            # oauthlib supports this. E.g., if access tokens are sent via URI
            # query parameters, such tokens may leak to log files and the HTTP
            # 'referer'.
            return False
        instance = Token.objects.lookup(Token.Type.ACCESS_TOKEN, token)
        if not instance:
            return False
        if instance.user and not instance.user.is_active:
            return False
        granted_scopes = instance.get_scopes()
        if not set(scopes).issubset(set(granted_scopes)):
            return False
        request.user = instance.user
        if not instance.client:
            return False
        ctx = get_validator_context()
        if (
            instance.client.type == Client.Type.PUBLIC
            and Client.AuthenticationMethod.NONE not in ctx.supported_auth_methods
        ):
            return False
        self._use_client(request, instance.client)
        request.scopes = granted_scopes
        ctx.access_token = instance
        request.access_token = token
        return True

    def revoke_token(self, token, token_type_hint, request, *args, **kwargs) -> None:
        if token_type_hint == "access_token":  # nosec
            types = [Token.Type.ACCESS_TOKEN]
        elif token_type_hint == "refresh_token":  # nosec
            types = [Token.Type.REFRESH_TOKEN]
        else:
            types = [Token.Type.ACCESS_TOKEN, Token.Type.REFRESH_TOKEN]
        Token.objects.by_value(token).filter(type__in=types).delete()

    def get_userinfo_claims(self, request: Request) -> dict:
        access_token = get_validator_context().access_token
        assert access_token  # nosec
        email = access_token.get_scope_email()
        return get_adapter().get_claims(
            "userinfo", request.user, request.client, request.scopes, email=email
        )

    def get_default_redirect_uri(
        self, client_id, request: Request, *args, **kwargs
    ) -> None:
        # https://openid.net/specs/openid-financial-api-part-1-1_0.html#section-5.2.2
        # 9. shall require the redirect_uri in the authorization request;
        # So, don't support a default.
        return None

    def validate_user(
        self, username, password, client, request: Request, *args, **kwargs
    ) -> bool:
        """
        Note that this bypasses MFA, which is why the password grant is not
        recommended and hence disabled. This could work:

            try:
                user = get_account_adapter().authenticate(
                    context.request, username=username, password=password
                )
            except ValidationError:
                return False
            else:
                if not user:
                    return False
                request.user = user
                return True
        """
        return False

    def validate_refresh_token(
        self, refresh_token, client, request: Request, *args, **kwargs
    ) -> bool:
        token = Token.objects.filter(client=client).lookup(
            Token.Type.REFRESH_TOKEN, refresh_token
        )
        if not token:
            return False
        if not token.user or not token.user.is_active:
            return False
        request.user = token.user
        get_validator_context().refresh_token = token
        return True

    def get_original_scopes(
        self, refresh_token, request: Request, *args, **kwargs
    ) -> list[str]:
        rt = get_validator_context().refresh_token
        assert rt is not None  # nosec
        return rt.get_scopes()

    def client_authentication_required(self, request: Request, *args, **kwargs) -> bool:
        """
        Determine if client authentication is required for current request.
        """
        if (
            Client.AuthenticationMethod.NONE
            not in get_validator_context().supported_auth_methods
        ):
            return True
        if request.client_id and request.client_secret:
            return True

        client = self._lookup_client(request, request.client_id)
        if client and client.type == Client.Type.PUBLIC:
            return False
        return super().client_authentication_required(request, *args, **kwargs)

    def _lookup_client(self, request: Request, client_id: str) -> Client | None:
        """
        In various places, oauthlib documents:

            Note, while not strictly necessary it can often be very convenient
            to set request.client to the client object associated with the
            given client_id.

        It's unclear though that if this is not explicitly stated, and, we still
        were to set request.client, whether that could have adverse side
        effects. So, don't assign request.client here.
        """
        cache = get_validator_context().clients
        if client_id in cache:
            client = cache[client_id]
        else:
            client = lookup_client(client_id)
            cache[client_id] = client
        return client

    def _use_client(self, request: Request, client: Client) -> None:
        request.client = client
        request.client.client_id = client.id

    def _lookup_authorization_code(
        self, request: Request, client_id: str, code: str
    ) -> dict | None:
        cache = get_validator_context().codes
        key = (client_id, code)
        if key in cache:
            authorization_code = cache[key]
        else:
            authorization_code = authorization_codes.lookup(client_id, code)
            cache[key] = authorization_code
        return authorization_code

    def is_origin_allowed(
        self, client_id, origin, request: Request, *args, **kwargs
    ) -> bool:
        client = self._lookup_client(request, client_id)
        return bool(
            client
            and is_origin_allowed(
                origin, client.get_cors_origins(), client.allow_uri_wildcards
            )
        )

    def rotate_refresh_token(self, request: Request) -> bool:
        return app_settings.ROTATE_REFRESH_TOKEN

    def validate_silent_login(self, request: Request) -> bool:
        if context.request.user.is_authenticated:
            request.user = context.request.user
            return True
        return False

    def validate_silent_authorization(self, request: Request) -> bool:
        granted_scopes = set()
        tokens = Token.objects.valid().filter(
            user=context.request.user,
            type__in=[Token.Type.REFRESH_TOKEN, Token.Type.ACCESS_TOKEN],
        )
        for token in tokens.iterator():
            granted_scopes.update(token.get_scopes())
        return set(request.scopes).issubset(granted_scopes)

    def validate_jwt_bearer_token(self, token, scopes, request: Request) -> bool:
        payload = decode_jwt_token(token, verify_iss=True, verify_exp=True)
        if not payload:
            return False
        token_use = payload.get("token_use")
        if token_use == "access":  # nosec
            return self._validate_jwt_bearer_access_token(
                token, scopes, request, payload
            )
        return self._validate_jwt_bearer_id_token(token, scopes, request, payload)

    def _validate_jwt_bearer_id_token(self, token, scopes, request, payload) -> bool:
        if scopes:
            # We don't have scope for the ID token
            return False
        client_id = payload.get("aud")
        return self.validate_client_id(client_id, request)

    def _validate_jwt_bearer_access_token(
        self, token, scopes, request, payload
    ) -> bool:
        client_id = payload.get("client_id")
        if not self.validate_client_id(client_id, request):
            return False
        return self.validate_bearer_token(token, scopes, request)



from __future__ import annotations

import re
from typing import Any

from django.conf import settings
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest

from allauth.core.exceptions import ImmediateHttpResponse
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter,
    OAuth2CallbackView,
    OAuth2LoginView,
)


class ShopifyOAuth2Adapter(OAuth2Adapter):
    provider_id = "shopify"
    scope_delimiter = ","

    def _shop_domain(self):
        shop = self.request.GET.get("shop", "")
        if "." not in shop:
            shop = f"{shop}.myshopify.com"
        # Ensure the provided hostname parameter is a valid hostname,
        # ends with myshopify.com, and does not contain characters
        # other than letters (a-z), numbers (0-9), dots, and hyphens.
        if not re.match(r"^[a-z0-9-]+\.myshopify\.com$", shop):
            raise ImmediateHttpResponse(
                HttpResponseBadRequest("Invalid `shop` parameter")
            )
        return shop

    def _shop_url(self, path):
        shop = self._shop_domain()
        return f"https://{shop}{path}"

    @property
    def access_token_url(self):
        return self._shop_url("/admin/oauth/access_token")

    @property
    def authorize_url(self):
        return self._shop_url("/admin/oauth/authorize")

    @property
    def profile_url(self):
        return self._shop_url("/admin/shop.json")

    def complete_login(self, request: HttpRequest, app, token, **kwargs):
        headers = {"X-Shopify-Access-Token": f"{token.token}"}
        with get_adapter().get_requests_session() as sess:
            response = sess.get(self.profile_url, headers=headers)
            extra_data = response.json()
        associated_user = kwargs["response"].get("associated_user")
        if associated_user:
            extra_data["associated_user"] = associated_user
        return self.get_provider().sociallogin_from_response(request, extra_data)


class ShopifyOAuth2LoginView(OAuth2LoginView):
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any):
        is_embedded = (
            getattr(settings, "SOCIALACCOUNT_PROVIDERS", {})
            .get("shopify", {})
            .get("IS_EMBEDDED", False)
        )
        if is_embedded:
            # TODO: This bypasses LOGIN_ON_GET, but:
            #
            #     The Embedded App SDK (EASDK) and backwards compatibility layer
            #     are being removed from Shopify on January 1, 2022.
            #
            # So this needs to be dropped/revisitted anyway.
            response = super().dispatch(request, *args, **kwargs)
            """
            Shopify embedded apps (that run within an iFrame) require a JS
            (not server) redirect for starting the oauth2 process.

            See Also:
            https://help.shopify.com/api/sdks/embedded-app-sdk/getting-started#oauth
            """
            js = (
                '<!DOCTYPE html><html><head><script type="text/javascript">'
                f'window.top.location.href = "{getattr(response, "url", "/")}";'
                "</script></head><body></body></html>"
            )
            response = HttpResponse(content=js)
            # Because this view will be within shopify's iframe
            response.xframe_options_exempt = True
            return response
        return super().dispatch(request, *args, **kwargs)


oauth2_login = ShopifyOAuth2LoginView.adapter_view(ShopifyOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(ShopifyOAuth2Adapter)



from __future__ import annotations

import hashlib
import uuid
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal

from django.contrib.auth import get_user_model
from django.contrib.auth.base_user import AbstractBaseUser
from django.core.exceptions import ImproperlyConfigured
from django.core.management.utils import get_random_secret_key
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from allauth.account.internal.userkit import (
    str_to_user_id,
    user_id_to_str,
    user_username,
)
from allauth.account.models import EmailAddress
from allauth.core.internal.adapter import BaseAdapter
from allauth.core.internal.cryptokit import generate_user_code
from allauth.idp.oidc import app_settings
from allauth.idp.oidc.internal.private_keys import filter_keys, pick_signing_key


if TYPE_CHECKING:
    from allauth.idp.oidc.models import Client, PrivateKey, Token


class DefaultOIDCAdapter(BaseAdapter):
    """The adapter class allows you to override various functionality of the
    ``allauth.idp.oidc`` app.  To do so, point ``settings.IDP_OIDC_ADAPTER`` to
    your own class that derives from ``DefaultOIDCAdapter`` and override the
    behavior by altering the implementation of the methods according to your own
    needs.
    """

    scope_display = {
        "openid": _("View your user ID"),
        "email": _("View your email address"),
        "profile": _("View your basic profile information"),
    }

    def generate_client_id(self) -> str:
        """
        The client ID to use for newly created clients.
        """
        return uuid.uuid4().hex

    def generate_client_secret(self) -> str:
        """
        The client secret to use for newly created clients.
        """
        return get_random_secret_key()

    def generate_user_code(self) -> str:
        return generate_user_code(**app_settings.USER_CODE_FORMAT)

    def hash_token(self, token: str) -> str:
        """
        We don't store tokens directly, only the hash of the token. This methods generates
        that hash.
        """
        return hashlib.sha256(token.encode("utf-8")).hexdigest()

    def get_issuer(self) -> str:
        """
        Returns the URL of the issuer.
        """
        return self.request.build_absolute_uri("/").rstrip("/")

    def populate_id_token(
        self,
        id_token: dict[str, Any],
        client: Client,
        scopes: Iterable[str],
        **kwargs: Any,
    ) -> None:
        """
        This method can be used to alter the ID token payload. It is already populated
        with basic values. Depending on the client and requested scopes, you can
        expose additional information here.
        """
        pass

    def populate_access_token(
        self,
        access_token: dict[str, Any],
        *,
        client: Client,
        scopes: Iterable[str],
        user: AbstractBaseUser,
        **kwargs: Any,
    ) -> None:
        """
        This method can be used to alter the JWT access token payload. It is already
        populated with basic values.
        """
        pass

    def get_claims(
        self,
        purpose: Literal["id_token", "userinfo"],
        user: AbstractBaseUser,
        client: Client,
        scopes: Iterable[str],
        email: str | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """
        Return the claims to be included in the ID token or userinfo response.
        """
        claims: dict[str, Any] = {"sub": self.get_user_sub(client, user)}
        if "email" in scopes:
            address: EmailAddress | None = None
            if email:
                try:
                    address = EmailAddress.objects.get_for_user(user, email)
                except EmailAddress.DoesNotExist:
                    pass
            else:
                address = EmailAddress.objects.get_primary(user)
            if address:
                claims.update(
                    {
                        "email": address.email,
                        "email_verified": address.verified,
                    }
                )
        if "profile" in scopes:
            if hasattr(user, "get_full_name"):
                full_name = user.get_full_name()
            else:
                full_name = ""
            last_name = getattr(user, "last_name", None)
            first_name = getattr(user, "first_name", None)
            username = user_username(user)
            profile_claims = {
                "name": full_name,
                "given_name": first_name,
                "family_name": last_name,
                "preferred_username": username,
            }
            for claim_key, claim_value in profile_claims.items():
                if claim_value:
                    claims[claim_key] = claim_value
        return claims

    def get_user_sub(self, client: Client, user: AbstractBaseUser) -> str:
        """
        Returns the "sub" (subject identifier) for the given user.
        """
        return user_id_to_str(user)

    def get_user_by_sub(self, client: Client, sub: str) -> AbstractBaseUser | None:
        """
        Looks up a user, given its subject identifier. Returns `None` if no
        such user was found.
        """
        try:
            pk = str_to_user_id(sub)
        except ValueError:
            return None
        user = get_user_model().objects.filter(pk=pk).first()
        if not user or not user.is_active:
            return None
        return user

    def validate_client_registration(
        self,
        *,
        client: Client,
        client_metadata: dict[str, Any],
        token: Token | None,
        bearer_token: str | None,
        **kwargs: Any,
    ) -> None:
        """
        This method is called after all builtin validation was successful,
        and just before the actual client is being created. To intervene, raise
        a ``ValidationError`` or an ``ImmediateHttpResponse``.

        ``client``: The ``Client`` instance that is about to be saved.
        ``client_metadata``: The raw JSON payload from the DCR request.
        ``token``: The ``Token`` instance corresponding to the initial access
            token, or ``None`` if no token was provided.
        ``bearer_token``: The raw bearer token string from the ``Authorization``
            header, or ``None`` if no token was provided.
        """
        pass

    def validate_resource_uris(self, *, uris: list[str], **kwargs: Any) -> None:
        """
        Allows for custom validation of resource URIs (RFC 8707).
        Throw a ``ValidationError`` to reject the resource.
        """
        pass

    def populate_server_metadata(self, data: dict[str, str | list[str]]) -> None:
        """
        Allows for customizing the ``/.well-known/openid-configuration``
        payload, as specified in `RFC 8414`_ (OAuth 2.0 Authorization Server
        Metadata).

        .. _RFC 8414: https://www.rfc-editor.org/info/rfc8414
        """
        pass

    def is_cimd_url_allowed(self, url: str) -> bool:
        """
        Determines whether the given CIMD (Client ID Metadata Document) URL is
        accepted as a ``client_id``.

        Override this method to restrict which clients can authenticate via CIMD,
        for example by maintaining a domain allowlist.  The default implementation
        accepts all URLs that pass structural validation.
        """
        return True

    def get_jwks_cache_control(self) -> int:
        """
        Returns the cache control value for the JWKS endpoint. The default
        implementation returns the value of the ``IDP_OIDC_JWKS_CACHE_CONTROL``
        setting, clamped so that clients refetch before the next key drops out
        of the key set (i.e. before the soonest ``expires_at``).  Override this
        method to provide a different cache control value, e.g. in case of a
        secret manager / vault is used.
        """
        cache_control = app_settings.JWKS_CACHE_CONTROL
        now = timezone.now()
        # Clamp against the same key set that is actually published (so a custom
        # ``list_private_keys()`` override cannot diverge from the cache window).
        for key in self.list_private_keys(is_active=True):
            if key.expires_at is not None:
                cache_control = min(
                    cache_control, int((key.expires_at - now).total_seconds())
                )
        return max(0, cache_control)

    def list_private_keys(
        self,
        *,
        did_activate: Literal[True] | None = None,
        is_active: Literal[True] | None = None,
    ) -> list[PrivateKey]:
        """
        Returns the configured private keys, optionally filtered.  Pass
        ``did_activate=True`` to exclude keys whose ``not_before`` lies in the
        future, and/or ``is_active=True`` to exclude keys past their
        ``expires_at``.  Used both for token verification and for serving
        ``.well-known/jwks.json``.
        """
        return filter_keys(
            app_settings.PRIVATE_KEYS, did_activate=did_activate, is_active=is_active
        )

    def get_signing_key(self) -> PrivateKey:
        """
        Returns the private key used for signing new tokens: the most recently
        issued key that has activated and not yet expired.  Raises
        ``ImproperlyConfigured`` if no such key is found.
        """
        key = pick_signing_key(
            self.list_private_keys(did_activate=True, is_active=True)
        )
        if key is None:
            raise ImproperlyConfigured("No active private key.")
        return key

    def populate_introspection_response(
        self, *, response: dict[str, Any], token: Token
    ) -> None:
        """
        This method can be used to add additional information to the introspection
        response for a given token. The default implementation does nothing.
        """
        return None

    def is_introspection_allowed(
        self,
        token: Token,
        *,
        caller_client: Client,
    ) -> bool:
        """
        This method can be used to add additional checks to determine if a token is
        valid in introspection responses and if the caller client is allowed to introspect it.
        The default implementation allows all introspection requests for active tokens,
        regardless of the caller client.

        ``caller_client``: The authenticated introspection caller client.
        """
        return True


def get_adapter() -> DefaultOIDCAdapter:
    return import_string(app_settings.ADAPTER)()



from __future__ import annotations

import json
from http import HTTPStatus
from typing import Any

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import AbstractBaseUser
from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import PermissionDenied, ValidationError
from django.core.signing import BadSignature, Signer
from django.db import transaction
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseForbidden,
    HttpResponseRedirect,
    JsonResponse,
)
from django.middleware.csrf import CsrfViewMiddleware
from django.shortcuts import render
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.http import urlencode
from django.views import View
from django.views.decorators.clickjacking import xframe_options_deny
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.edit import FormView

from oauthlib.common import Request
from oauthlib.oauth2.rfc6749 import errors
from oauthlib.oauth2.rfc6749.errors import InvalidScopeError, OAuth2Error

from allauth.account import app_settings as account_settings
from allauth.account.adapter import get_adapter as get_account_adapter
from allauth.account.internal.decorators import login_not_required
from allauth.core.exceptions import ImmediateHttpResponse, RateLimited
from allauth.core.internal import jwkkit, ratelimit
from allauth.core.internal.httpkit import (
    add_query_params,
    authenticated_user,
    del_query_params,
)
from allauth.idp.oidc import app_settings
from allauth.idp.oidc.adapter import get_adapter
from allauth.idp.oidc.forms import (
    AuthorizationForm,
    ClientRegistrationForm,
    ConfirmCodeForm,
    DeviceAuthorizationForm,
    RPInitiatedLogoutForm,
)
from allauth.idp.oidc.internal import flows
from allauth.idp.oidc.internal.oauthlib import device_codes
from allauth.idp.oidc.internal.oauthlib.server import get_device_server, get_server
from allauth.idp.oidc.internal.oauthlib.utils import (
    convert_response,
    extract_params,
    get_validator_context,
    respond_html_error,
    respond_json_error,
)
from allauth.idp.oidc.internal.resources import get_resources
from allauth.idp.oidc.models import Client, Token
from allauth.utils import build_absolute_uri


class _ClientHttpResponseRedirect(HttpResponse):
    """
    Not using HttpResponseRedirect -- that one checks schemes, and here
    we might be redirecting to an app native url.
    """

    def __init__(self, redirect_uri: str):
        super().__init__(status=HTTPStatus.FOUND)
        self["Location"] = redirect_uri


def _enforce_csrf(request: HttpRequest) -> HttpResponseForbidden | None:
    """
    Scenario: view is CSRF exempt, but, if this is not a client initial POST
    request, we do want a properly CSRF protected view.
    """
    reason = CsrfViewMiddleware(
        get_response=lambda req: HttpResponseForbidden()
    ).process_view(request, lambda *args, **kwargs: HttpResponse(), (), {})
    if reason:
        return HttpResponseForbidden(f"CSRF Failed: {reason}")
    return None


@method_decorator(login_not_required, name="dispatch")
class ConfigurationView(View):
    def get(self, request: HttpRequest) -> JsonResponse:
        userinfo_endpoint = app_settings.USERINFO_ENDPOINT
        if not userinfo_endpoint:
            userinfo_endpoint = build_absolute_uri(
                request, reverse("idp:oidc:userinfo")
            )
        supported_types = self._get_supported_types()
        data = {
            "authorization_endpoint": build_absolute_uri(
                request, reverse("idp:oidc:authorization")
            ),
            "code_challenge_methods_supported": ["S256"],
            "device_authorization_endpoint": build_absolute_uri(
                request, reverse("idp:oidc:device_code")
            ),
            "end_session_endpoint": build_absolute_uri(
                request, reverse("idp:oidc:logout")
            ),
            "id_token_signing_alg_values_supported": ["RS256"],
            "issuer": get_adapter().get_issuer(),
            "jwks_uri": build_absolute_uri(request, reverse("idp:oidc:jwks")),
            "revocation_endpoint": build_absolute_uri(
                request, reverse("idp:oidc:revoke")
            ),
            "token_endpoint": build_absolute_uri(request, reverse("idp:oidc:token")),
            "token_endpoint_auth_methods_supported": app_settings.AUTH_METHODS,
            "userinfo_endpoint": userinfo_endpoint,
            "subject_types_supported": ["public"],
            **supported_types,
        }
        if app_settings.INTROSPECTION_ENABLED:
            data["introspection_endpoint"] = build_absolute_uri(
                request, reverse("idp:oidc:introspect")
            )
            data["introspection_endpoint_auth_methods_supported"] = (
                app_settings.INTROSPECTION_AUTH_METHODS
            )
        if app_settings.DCR_ENABLED:
            data["registration_endpoint"] = build_absolute_uri(
                request, reverse("idp:oidc:client_registration")
            )

        get_adapter().populate_server_metadata(data)
        response = JsonResponse(data)
        response["Access-Control-Allow-Origin"] = "*"
        return response

    def _get_supported_types(self) -> dict[str, list[str]]:
        scopes_supported = ["openid", "profile", "email"]
        return {
            "scopes_supported": scopes_supported,
            "grant_types_supported": sorted(gt.value for gt in Client.GrantType),
            "response_types_supported": sorted(rt.value for rt in Client.ResponseType),
        }


configuration = ConfigurationView.as_view()


@method_decorator(xframe_options_deny, name="dispatch")
@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class AuthorizationView(FormView):
    form_class = AuthorizationForm
    template_name = f"idp/oidc/authorization_form.{account_settings.TEMPLATE_EXTENSION}"

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        response = self._login_required(request)
        if response:
            return response
        orequest = extract_params(self.request)
        try:
            server = get_server()
            self._scopes, self._request_info = server.validate_authorization_request(
                *orequest
            )
            self._request_info["resources"] = get_resources(request)
            if "none" in self._request_info.get("prompt", ()):
                oresponse = server.create_authorization_response(
                    *orequest, scopes=self._scopes
                )
                return convert_response(*oresponse)

        # Errors that should be shown to the user on the provider website
        except (errors.FatalClientError, ValidationError) as e:
            return respond_html_error(request, error=e)
        except errors.OAuth2Error as e:
            return HttpResponseRedirect(e.in_uri(e.redirect_uri))
        if self._request_info["request"].client.skip_consent:
            return self._skip_consent()
        return super().get(request, *args, **kwargs)

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        signed_request_info = request.POST.get("request")
        if not signed_request_info:
            return HttpResponseRedirect(
                f"{reverse('idp:oidc:authorization')}?{request.POST.urlencode()}"
            )
        response = self._login_required(request)
        if response:
            return response

        csrf_resp = _enforce_csrf(request)
        if csrf_resp:
            return csrf_resp

        try:
            signer = Signer()
            self._scopes, self._request_info = signer.unsign_object(signed_request_info)
        except BadSignature:
            raise PermissionDenied
        if request.POST.get("action") != "grant":
            return self._respond_with_access_denied()
        return super().post(request, *args, **kwargs)

    def _login_required(self, request: HttpRequest) -> HttpResponse | None:
        prompts = []
        prompt = request.GET.get("prompt")
        if prompt:
            prompts = prompt.split()
        if "login" in prompts:
            return self._handle_login_prompt(request, prompts)
        if "none" in prompts:
            return None
        if request.user.is_authenticated:
            return None
        return login_required()(None)(request)  # type: ignore[misc,type-var]

    def _handle_login_prompt(
        self, request: HttpRequest, prompts: list[str]
    ) -> HttpResponse:
        prompts.remove("login")
        next_url = request.get_full_path()
        if prompts:
            next_url = add_query_params(next_url, {"prompt": " ".join(prompts)})
        else:
            next_url = del_query_params(next_url, "prompt")
        params = {}
        params[REDIRECT_FIELD_NAME] = next_url
        path = reverse(
            "account_reauthenticate"
            if request.user.is_authenticated
            else "account_login"
        )
        return HttpResponseRedirect(add_query_params(path, params))

    def _skip_consent(self) -> HttpResponse:
        scopes = self._request_info["request"].scopes
        form_kwargs = self.get_form_kwargs()
        form_kwargs["data"] = {
            "scopes": scopes,
            "request": "not-relevant-for-skip-consent",
        }
        form = self.form_class(**form_kwargs)
        if not form.is_valid():
            # Shouldn't occur.
            raise PermissionDenied()
        return self.form_valid(form)

    def _respond_with_access_denied(self) -> HttpResponse:
        redirect_uri = self._request_info.get("redirect_uri")
        state = self._request_info.get("state")
        params = {"error": "access_denied"}
        if state:
            params["state"] = state
        return _ClientHttpResponseRedirect(add_query_params(redirect_uri, params))

    def get_form_kwargs(self) -> dict[str, Any]:
        ret = super().get_form_kwargs()
        ret.update({"requested_scopes": self._scopes, "user": self.request.user})
        return ret

    def get_initial(self) -> dict[str, Any]:
        signer = Signer()
        ret = {}
        request_info = self._request_info
        request_info.pop("request", None)
        prompt = request_info.get("prompt")
        if isinstance(prompt, set):
            request_info["prompt"] = list(prompt)
        ret["request"] = signer.sign_object((self._scopes, request_info))
        return ret

    def form_valid(self, form: AuthorizationForm) -> HttpResponse:
        orequest = extract_params(self.request)

        scopes = form.cleaned_data["scopes"]

        # oauthlib puts all credentials into its `Request`.
        credentials = {"user": self.request.user}
        credentials.update(self._request_info)
        credentials.pop("resources", None)

        ctx = get_validator_context()
        ctx.requested_resources = self._request_info.get("resources")

        try:
            email = form.cleaned_data.get("email")
            if email:
                ctx.email = email
            oresponse = get_server().create_authorization_response(
                *orequest, scopes=scopes, credentials=credentials
            )
            return convert_response(*oresponse)

        except errors.FatalClientError as e:
            return respond_html_error(self.request, error=e)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ret = super().get_context_data(**kwargs)
        ret.update(
            {
                "client": Client.objects.get(id=self._request_info["client_id"]),
                "site": get_current_site(self.request),
            }
        )
        return ret


authorization = AuthorizationView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class DeviceCodeView(View):
    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        orequest = extract_params(request)
        try:
            get_validator_context().requested_resources = get_resources(request)
            headers, data, status = (
                get_device_server().create_device_authorization_response(*orequest)
            )
            if status == HTTPStatus.OK:
                client_id = request.POST["client_id"]
                scope: list[str] | None = None
                if "scope" in request.POST:
                    scope = request.POST["scope"].split()
                    client = Client.objects.get(id=client_id)
                    if not set(scope).issubset(set(client.get_scopes())):
                        raise InvalidScopeError()
                device_codes.create(client_id, scope, data)
        except OAuth2Error as e:
            return HttpResponse(
                e.json, content_type="application/json", status=e.status_code
            )
        return convert_response(headers, data, status)


device_code = DeviceCodeView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_required, name="dispatch")
class DeviceAuthorizationView(View):
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if "code" in request.GET:
            form = ConfirmCodeForm(request.GET)
            if form.is_valid():
                return self._dispatch_authorization(
                    request,
                    form.cleaned_data["code"],
                    form.device_code,
                    form.client,
                )
        else:
            form = ConfirmCodeForm()
        context = {
            "form": form,
            "autorization_url": reverse("idp:oidc:device_authorization"),
        }
        return render(
            request,
            f"idp/oidc/device_authorization_code_form.{account_settings.TEMPLATE_EXTENSION}",
            context,
        )

    def _dispatch_authorization(
        self, request: HttpRequest, user_code: str, device_code: str, client: Client
    ) -> HttpResponse:
        context = {"user_code": user_code, "client": client}
        if request.method == "POST":
            form = DeviceAuthorizationForm(request.POST)
            if form.is_valid():
                confirm = form.cleaned_data["action"] == "confirm"
                device_codes.confirm_or_deny_device_code(
                    authenticated_user(request), device_code, confirm=confirm
                )
                if confirm:
                    template_name = f"idp/oidc/device_authorization_confirmed.{account_settings.TEMPLATE_EXTENSION}"
                else:
                    template_name = f"idp/oidc/device_authorization_denied.{account_settings.TEMPLATE_EXTENSION}"
                return render(request, template_name, context)
        else:
            form = DeviceAuthorizationForm()
        context["autorization_url"] = (
            reverse("idp:oidc:device_authorization")
            + "?"
            + urlencode({"code": user_code})
        )

        return render(
            request,
            f"idp/oidc/device_authorization_confirm_form.{account_settings.TEMPLATE_EXTENSION}",
            context,
        )


device_authorization = DeviceAuthorizationView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class TokenView(View):

    @transaction.atomic
    def post(self, request: HttpRequest) -> HttpResponse:
        try:
            get_validator_context().requested_resources = get_resources(request)
            if request.POST.get("grant_type") == Client.GrantType.DEVICE_CODE:
                return self._post_device_token(request)
            return self._create_token_response(request)
        except OAuth2Error as e:
            resp = JsonResponse(dict(e.twotuples))
            resp.status_code = e.status_code
            return resp

    def _create_token_response(
        self,
        request: HttpRequest,
        *,
        user: AbstractBaseUser | None = None,
        data: dict[str, Any] | None = None,
    ) -> HttpResponse:
        orequest = extract_params(request)
        oresponse = get_server(
            pre_token=[lambda orequest: self._pre_token(orequest, user, data)],
        ).create_token_response(*orequest)
        return convert_response(*oresponse)

    def _pre_token(
        self,
        orequest: Request,
        user: AbstractBaseUser | None,
        data: dict[str, Any] | None,
    ) -> None:
        if orequest.grant_type == Client.GrantType.DEVICE_CODE:
            assert user is not None  # nosec
            assert data is not None  # nosec
            if scope := data.get("scope"):
                orequest.scope = scope
            orequest.user = user

    def _post_device_token(self, request: HttpRequest) -> HttpResponse:
        try:
            user, data = device_codes.poll_device_code(request)
        except OAuth2Error as e:
            return HttpResponse(
                e.json, content_type="application/json", status=e.status_code
            )
        else:
            return self._create_token_response(request, user=user, data=data)


token = TokenView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class UserInfoView(View):
    """
    The UserInfo Endpoint MUST support the use of the HTTP GET and HTTP POST methods
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        return self._respond(request)

    def post(self, request: HttpRequest) -> HttpResponse:
        return self._respond(request)

    def _respond(self, request: HttpRequest) -> HttpResponse:
        orequest = extract_params(request)
        try:
            oresponse = get_server().create_userinfo_response(*orequest)
            return convert_response(*oresponse)
        except OAuth2Error as e:
            return respond_json_error(request, e)


user_info = UserInfoView.as_view()


@method_decorator(login_not_required, name="dispatch")
class JwksView(View):
    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> JsonResponse:
        keys = []
        adapter = get_adapter()
        # Deliberately not filtering on ``did_activate``: keys whose
        # ``not_before`` still lies in the future are published ahead of time.
        # That way clients have already fetched and cached the next key before
        # it starts signing, avoiding a window where a freshly activated key
        # signs tokens that verifiers cannot yet validate.
        for key in adapter.list_private_keys(is_active=True):
            jwk, _ = jwkkit.load_jwk_from_pem(key.pem)
            keys.append(jwk)
        response = JsonResponse({"keys": keys})
        response["Access-Control-Allow-Origin"] = "*"
        response["Cache-Control"] = (
            f"max-age={adapter.get_jwks_cache_control()}, must-revalidate"
        )

        return response


jwks = JwksView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class RevokeView(View):
    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        orequest = extract_params(request)
        oresponse = get_server().create_revocation_response(*orequest)
        return convert_response(*oresponse)


revoke = RevokeView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class IntrospectView(View):
    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        ctx = get_validator_context()
        ctx.supported_auth_methods = app_settings.INTROSPECTION_AUTH_METHODS
        orequest = extract_params(request)
        try:
            # Throttle by source IP before client authentication runs, so that
            # unauthenticated and failed-authentication requests are bounded as
            # well. The per-client limit is consumed later, post-authentication.
            ratelimit.consume(
                request=request,
                config=app_settings.RATE_LIMITS,
                action="introspect_ip",
                raise_exception=True,
            )
            oresponse = get_server().create_introspect_response(*orequest)
        except RateLimited:
            return JsonResponse(
                {
                    "error": "temporarily_unavailable",
                    "error_description": "Too many requests.",
                },
                status=HTTPStatus.TOO_MANY_REQUESTS,
            )
        return convert_response(*oresponse)


introspect = IntrospectView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class LogoutView(FormView):
    """
    https://openid.net/specs/openid-connect-rpinitiated-1_0.html
    """

    form_class = RPInitiatedLogoutForm
    template_name = f"idp/oidc/logout.{account_settings.TEMPLATE_EXTENSION}"

    def get(self, request: HttpRequest) -> HttpResponse:
        form = self.form_class(request.GET)
        if not form.is_valid():
            return self.form_invalid(form)
        if not self._must_ask(form):
            return self._handle(form, True)
        return self.render_to_response(self.get_context_data(form=form))

    def form_invalid(self, form: RPInitiatedLogoutForm) -> HttpResponse:
        return respond_html_error(self.request, form=form)

    def form_valid(self, form: RPInitiatedLogoutForm) -> HttpResponse:
        ask = self._must_ask(form)
        action = form.cleaned_data["action"]
        if ask:
            # If we're supposed to ask, we need to ensure this POST request does
            # NOT come from the RP, but from the actual user visitting the logout
            # page.
            csrf_token = self.request.POST.get("csrfmiddlewaretoken", "")
            if not csrf_token or not action:
                return self.render_to_response(self.get_context_data(form=form))
            csrf_resp = _enforce_csrf(self.request)
            if csrf_resp:
                return csrf_resp
            op_logout = action != "stay"
        else:
            op_logout = True
        return self._handle(form, op_logout)

    def _handle(self, form: RPInitiatedLogoutForm, op_logout: bool) -> HttpResponse:
        cleaned_data = form.cleaned_data
        flows.rp_initiated_logout(
            self.request,
            from_op=op_logout,
            client=cleaned_data["client"],
            post_logout_redirect_uri=cleaned_data["post_logout_redirect_uri"],
        )
        redirect_uri = cleaned_data["post_logout_redirect_uri"]
        if redirect_uri:
            state = cleaned_data["state"]
            if state:
                redirect_uri = add_query_params(redirect_uri, {"state": state})
        else:
            redirect_uri = get_account_adapter().get_logout_redirect_url(self.request)
        return _ClientHttpResponseRedirect(redirect_uri)

    def _must_ask(self, form: RPInitiatedLogoutForm) -> bool:
        """
        At the Logout Endpoint, the OP SHOULD ask the End-User whether to
        log out of the OP as well. Furthermore, the OP MUST ask the End-User
        this question if an id_token_hint was not provided or if the supplied ID
        Token does not belong to the current OP session with the RP and/or
        currently logged in End-User. If the End-User says "yes", then the OP
        MUST log out the End-User.
        """
        if self.request.user.is_anonymous:
            return False
        if app_settings.RP_INITIATED_LOGOUT_ASKS_FOR_OP_LOGOUT:
            return True
        id_token_hint = form.cleaned_data["id_token_hint"]
        sub = None
        if id_token_hint:
            sub = id_token_hint.get("sub")
        client = form.cleaned_data.get("client")
        if not id_token_hint or not client or not sub:
            return True
        user_hint = get_adapter().get_user_by_sub(client, sub)
        if not user_hint or (user_hint.pk != self.request.user.pk):
            return True
        return False


logout = LogoutView.as_view()


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(login_not_required, name="dispatch")
class ClientRegistrationView(View):

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        bearer_token, token, resp = self._authorize()
        if resp:
            return resp

        client_metadata = self._get_client_metadata()
        if client_metadata is None:
            return JsonResponse(
                {
                    "error": "invalid_client_metadata",
                    "error_description": "Invalid JSON data.",
                },
                status=HTTPStatus.BAD_REQUEST,
            )
        form = ClientRegistrationForm(data=client_metadata)
        if not form.is_valid():
            return self._form_invalid(form)
        return self._form_valid(
            form,
            client_metadata=client_metadata,
            bearer_token=bearer_token,
            token=token,
        )

    @transaction.atomic
    def _form_valid(
        self,
        form: ClientRegistrationForm,
        *,
        client_metadata: dict[str, Any],
        token: Token | None,
        bearer_token: str | None,
    ) -> HttpResponse:
        client = form.save(commit=False)
        client.data = {
            "dcr": True,
            "client_metadata": client_metadata,
        }
        secret = get_adapter().generate_client_secret()
        client.set_secret(secret)

        resp: HttpResponse | None
        usage, resp = self._ratelimit()
        if resp:
            return resp
        assert usage  # nosec
        resp = self._perform_custom_validation(
            client=client,
            client_metadata=client_metadata,
            token=token,
            bearer_token=bearer_token,
        )
        if resp:
            usage.rollback()
            return resp

        client.save()
        return JsonResponse(
            self._serialize_client(client, form, secret), status=HTTPStatus.CREATED
        )

    def _form_invalid(self, form: ClientRegistrationForm) -> JsonResponse:
        error_fields = list(form.errors.keys())
        if form.data and error_fields == ["redirect_uris"]:
            error = "invalid_redirect_uri"
        else:
            error = "invalid_client_metadata"

        if not form.data:
            error_description = "Invalid data received."
        else:
            error_description = "  ".join(
                [
                    f"'{field}': {' '.join([str(err) for err in errs])}"
                    for field, errs in form.errors.items()
                ]
            )
        return JsonResponse(
            {
                "error": error,
                "error_description": error_description,
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    def _authorize(self) -> tuple[str | None, Token | None, HttpResponse | None]:
        if not app_settings.DCR_REQUIRES_INITIAL_ACCESS_TOKEN:
            return None, None, None
        auth = self.request.headers.get("Authorization", "")
        scheme, _, bearer_token = auth.partition(" ")
        if scheme.lower() != "bearer" or not bearer_token:
            return None, None, self._respond_unauthorized()
        token = Token.objects.lookup(Token.Type.INITIAL_ACCESS_TOKEN, bearer_token)
        if not token:
            return None, None, self._respond_unauthorized()
        return bearer_token, token, None

    def _respond_unauthorized(self) -> HttpResponse:
        resp = HttpResponse(status=HTTPStatus.UNAUTHORIZED)
        resp["WWW-Authenticate"] = 'Bearer error="invalid_token"'
        return resp

    def _get_client_metadata(self) -> dict[str, Any] | None:
        try:
            data = json.loads(self.request.body)
        except (json.JSONDecodeError, ValueError):
            return None
        if not isinstance(data, dict):
            return None
        return data

    def _ratelimit(self) -> tuple[ratelimit.RateLimitUsage | None, JsonResponse | None]:
        usage = ratelimit.consume(
            request=self.request,
            config=app_settings.RATE_LIMITS,
            action="client_registration",
            raise_exception=False,
        )
        resp: JsonResponse | None = None
        if not usage:
            resp = JsonResponse(
                {
                    "error": "temporarily_unavailable",
                    "error_description": "Too many requests.",
                },
                status=HTTPStatus.TOO_MANY_REQUESTS,
            )
        return usage, resp

    def _perform_custom_validation(
        self,
        *,
        client: Client,
        client_metadata: dict[str, Any],
        token: Token | None,
        bearer_token: str | None,
    ) -> HttpResponse | None:
        try:
            get_adapter().validate_client_registration(
                client=client,
                client_metadata=client_metadata,
                token=token,
                bearer_token=bearer_token,
            )
        except ValidationError as e:
            return JsonResponse(
                {
                    "error": "invalid_client_metadata",
                    "error_description": str(e.message),
                },
                status=HTTPStatus.BAD_REQUEST,
            )
        except ImmediateHttpResponse as e:
            return e.response
        return None

    def _serialize_client(
        self, client: Client, form: ClientRegistrationForm, secret: str
    ) -> dict[str, Any]:
        data: dict[str, Any] = {
            "client_id": client.id,
            "client_name": client.name,
            "token_endpoint_auth_method": form.cleaned_data[
                "token_endpoint_auth_method"
            ],
            "scope": " ".join(client.get_scopes()),
            "client_id_issued_at": int(client.created_at.timestamp()),
            "redirect_uris": client.get_redirect_uris(),
            "grant_types": client.get_grant_types(),
            "response_types": client.get_response_types(),
        }
        if client.type == Client.Type.CONFIDENTIAL:
            data.update(
                {
                    "client_secret": secret,
                    "client_secret_expires_at": 0,  # nosec
                }
            )
        return data


client_registration = ClientRegistrationView.as_view()



from __future__ import annotations

from datetime import datetime, timezone as dt_timezone
from typing import TYPE_CHECKING, Any, TypeVar

from django.core.exceptions import ImproperlyConfigured

from allauth import app_settings as allauth_settings
from allauth.core.internal.cryptokit import UserCodeFormat


if TYPE_CHECKING:
    from allauth.idp.oidc.models import PrivateKey

_T = TypeVar("_T")


class AppSettings:
    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    def _setting(self, name: str, dflt: _T) -> _T:
        from allauth.utils import get_setting

        return get_setting(f"{self.prefix}{name}", dflt)

    @property
    def ADAPTER(self) -> str:
        return self._setting(
            "ADAPTER",
            "allauth.idp.oidc.adapter.DefaultOIDCAdapter",
        )

    @property
    def ID_TOKEN_EXPIRES_IN(self) -> int:
        return self._setting("ID_TOKEN_EXPIRES_IN", 5 * 60)

    @property
    def PRIVATE_KEY(self) -> str:
        return self._setting("PRIVATE_KEY", "")

    @property
    def PRIVATE_KEYS(self) -> list[PrivateKey]:
        from allauth.idp.oidc.models import PrivateKey

        ret: list[PrivateKey] = []
        keys: Any = self._setting("PRIVATE_KEYS", [])
        if not isinstance(keys, (list, tuple)):
            raise ImproperlyConfigured("IDP_OIDC_PRIVATE_KEYS: must be a list")
        pem1 = self._setting("PRIVATE_KEY", "")
        seen_pems: set[str] = set()
        for key in keys:
            if not isinstance(key, dict):
                raise ImproperlyConfigured(
                    "IDP_OIDC_PRIVATE_KEYS: each entry must be a dict"
                )
            pem = key.get("pem")
            if not isinstance(pem, str):
                raise ImproperlyConfigured(
                    "IDP_OIDC_PRIVATE_KEYS: 'pem' must be a string"
                )
            if pem in seen_pems:
                raise ImproperlyConfigured("IDP_OIDC_PRIVATE_KEYS: duplicate 'pem'")
            seen_pems.add(pem)
            ret.append(
                PrivateKey(
                    pem=pem,
                    not_before=self._parse_datetime(
                        "not_before", key.get("not_before")
                    ),
                    expires_at=self._parse_datetime(
                        "expires_at", key.get("expires_at")
                    ),
                    issued_at=self._parse_datetime("issued_at", key.get("issued_at")),
                )
            )
        if pem1 and pem1 not in seen_pems:
            ret.append(PrivateKey(pem=pem1))
        return ret

    @property
    def JWKS_CACHE_CONTROL(self) -> int:
        return self._setting("JWKS_CACHE_CONTROL", 60 * 60)

    @property
    def ACCESS_TOKEN_EXPIRES_IN(self) -> int:
        return self._setting("ACCESS_TOKEN_EXPIRES_IN", 3600)

    @property
    def ACCESS_TOKEN_FORMAT(self) -> str:
        return self._setting("ACCESS_TOKEN_FORMAT", "opaque")

    @property
    def AUTH_METHODS(self) -> tuple[str, ...]:
        return tuple(
            self._setting(
                "AUTH_METHODS",
                ["client_secret_basic", "client_secret_post", "none"],
            )
        )

    @property
    def INTROSPECTION_ENABLED(self) -> bool:
        return self._setting("INTROSPECTION_ENABLED", False)

    @property
    def INTROSPECTION_AUTH_METHODS(self) -> tuple[str, ...]:
        return tuple(
            self._setting(
                "INTROSPECTION_AUTH_METHODS",
                ["client_secret_basic", "client_secret_post"],
            )
        )

    @property
    def INTROSPECTION_CROSS_CLIENT_ALLOWED(self) -> bool:
        return self._setting("INTROSPECTION_CROSS_CLIENT_ALLOWED", False)

    @property
    def AUTHORIZATION_CODE_EXPIRES_IN(self) -> int:
        return self._setting("AUTHORIZATION_CODE_EXPIRES_IN", 60)

    @property
    def ROTATE_REFRESH_TOKEN(self) -> bool:
        return self._setting("ROTATE_REFRESH_TOKEN", True)

    @property
    def REFRESH_TOKEN_EXPIRES_IN(self) -> int | None:
        return self._setting("REFRESH_TOKEN_EXPIRES_IN", None)

    @property
    def DEVICE_CODE_EXPIRES_IN(self) -> int:
        return self._setting("DEVICE_CODE_EXPIRES_IN", 300)

    @property
    def DEVICE_CODE_INTERVAL(self) -> int:
        return self._setting("DEVICE_CODE_INTERVAL", 5)

    @property
    def USER_CODE_FORMAT(self) -> UserCodeFormat:
        return self._setting("USER_CODE_FORMAT", allauth_settings.USER_CODE_FORMAT)

    @property
    def RATE_LIMITS(self) -> dict:
        rls: dict = self._setting("RATE_LIMITS", {})
        if rls is False:
            return {}
        ret = {
            # OIDC device user code checks
            "device_user_code": "5/m/ip",
            # DCR
            "client_registration": "3/m/ip",
            # CIMD fetches
            "cimd_fetch": "3/m/ip",
            # token introspection, throttled per source IP (before client
            # authentication) and per authenticated client (after).
            "introspect_ip": "30/m/ip",
            "introspect_client": "60/m/key",
        }
        ret.update(rls)
        return ret

    @property
    def RP_INITIATED_LOGOUT_ASKS_FOR_OP_LOGOUT(self) -> bool:
        """
        At https://openid.net/specs/openid-connect-rpinitiated-1_0.html

        > 2. RP-Initiated Logout':
        > At the Logout Endpoint, the OP SHOULD ask the End-User whether to
        > log out of the OP as well.

        This setting controls whether the OP always asks.
        """
        return self._setting("RP_INITIATED_LOGOUT_ASKS_FOR_OP_LOGOUT", True)

    @property
    def USERINFO_ENDPOINT(self) -> str | None:
        """
        This setting can be used to point the ``userinfo_endpoint`` value as
        returned in the ".well-known/openid-configuration" to a custom URL.
        Setting this disables the built-in userinfo endpoint.
        """
        return self._setting("USERINFO_ENDPOINT", None)

    @property
    def DCR_ENABLED(self) -> bool:
        """
        Controls whether Dynamic Client Registration (RFC 7591) is enabled.
        """
        return self._setting("DCR_ENABLED", False)

    @property
    def DCR_REQUIRES_INITIAL_ACCESS_TOKEN(self) -> bool:
        """
        Controls whether or not an initial access token is required to
        perform Dynamic Client Registration.
        """
        return self._setting("DCR_REQUIRES_INITIAL_ACCESS_TOKEN", True)

    @property
    def CIMD_ENABLED(self) -> bool:
        """
        Whether or not CIMD is enabled.

        https://datatracker.ietf.org/doc/draft-ietf-oauth-client-id-metadata-document/
        https://client.dev/
        """
        return self._setting("CIMD_ENABLED", False)

    @property
    def CIMD_CACHE_TIMEOUT(self) -> int:
        """
        CIMD cache duration, in seconds.
        """
        return self._setting("CIMD_CACHE_TIMEOUT", 60 * 60)

    @staticmethod
    def _parse_datetime(field: str, value: Any) -> datetime | None:
        if value is None:
            return None
        dt: datetime | None = None
        if isinstance(value, str):
            try:
                dt = datetime.fromisoformat(value)
            except ValueError:
                pass
        elif isinstance(value, datetime):
            dt = value
        if dt is None:
            raise ImproperlyConfigured(
                f"IDP_OIDC_PRIVATE_KEYS: {field!r} is not a valid datetime: {value!r}"
            )
        if dt.tzinfo is None:
            return dt.replace(tzinfo=dt_timezone.utc)
        return dt.astimezone(dt_timezone.utc)


_app_settings = AppSettings("IDP_OIDC_")


def __getattr__(name):
    # See https://peps.python.org/pep-0562/
    return getattr(_app_settings, name)



from __future__ import annotations

from typing import Any

from allauth.core.internal import jwkkit
from allauth.core.internal.deferred import jwt
from allauth.idp.oidc.adapter import get_adapter
from allauth.idp.oidc.models import Token


def decode_jwt_token(
    value: str, *, client_id: str | None = None, verify_exp: bool, verify_iss: bool
) -> dict[str, Any] | None:
    if not value:
        return None
    try:
        headers = jwt.get_unverified_header(value)

        if "kid" not in headers:
            return None

        adapter = get_adapter()
        for key in adapter.list_private_keys(is_active=True):
            jwk_dict, private_key = jwkkit.load_jwk_from_pem(key.pem)
            if jwk_dict["kid"] == headers["kid"]:
                break
        else:
            return None

        issuer: str | None = None
        audience: str | None = None
        if client_id:
            audience = client_id
        if verify_iss:
            issuer = adapter.get_issuer()
        return jwt.decode(
            value,
            key=private_key.public_key(),
            algorithms=["RS256"],
            options={
                "verify_signature": True,
                "verify_iss": verify_iss,
                "verify_aud": client_id is not None,
                "verify_exp": verify_exp,
            },
            audience=audience,
            issuer=issuer,
        )
    except jwt.PyJWTError:
        return None


def is_jwt_token(token: str) -> bool:
    # Same check as done by the `JWTToken` class of `oauthlib`
    # count == "2" should be sufficient as currently only JWS as AccessTokens is supported,
    # but count == "4" allows also for JWE, if that is added in the future.
    # This is more of a "loose" check, but as those JWT that aren't detected using this check
    # will be rejected by the verify_request anyways, we can use the same logic here.
    return (
        isinstance(token, str) and token.startswith("ey") and token.count(".") in (2, 4)
    )


def determine_token_type(
    token: str, token_type_hint: str | None
) -> tuple[list[Token.Type] | None, list[Token.Type] | None]:
    """
    https://datatracker.ietf.org/doc/html/rfc7009#section-2.1
    > token_type_hint  OPTIONAL.  A hint about the type of the token
    >   submitted for revocation.  Clients MAY pass this parameter in
    >   order to help the authorization server to optimize the token
    >   lookup.  If the server is unable to locate the token using
    >   the given hint, it MUST extend its search across all of its
    >   supported token types.  An authorization server MAY ignore
    >   this parameter, particularly if it is able to detect the
    >   token type automatically.  This specification defines two
    >   such values:

    >   * access_token: An access token as defined in [RFC6749],
    >     Section 1.4

    >   * refresh_token: A refresh token as defined in [RFC6749],
    >     Section 1.5

    >   Specific implementations, profiles, and extensions of this
    >   specification MAY define other values for this parameter
    >   using the registry defined in Section 4.1.2.

    Returns a tuple of (token_types_to_try, token_type_for_error_response)
    """

    # This allows us to still handle JWT tokens when the configured access token format has been changed
    if is_jwt_token(token):
        payload = decode_jwt_token(token, verify_exp=True, verify_iss=True)
        if payload:
            return (
                (
                    [Token.Type.ACCESS_TOKEN]
                    if payload.get("token_use") == "access"
                    else None
                ),
                None,
            )
        # In theory, an opaque token could start with "ey...", so continue.
    default = ([Token.Type.ACCESS_TOKEN, Token.Type.REFRESH_TOKEN], None)
    if not isinstance(token_type_hint, str):
        return default
    return {
        "access_token": ([Token.Type.ACCESS_TOKEN], [Token.Type.REFRESH_TOKEN]),
        "refresh_token": ([Token.Type.REFRESH_TOKEN], [Token.Type.ACCESS_TOKEN]),
    }.get(token_type_hint, default)



from __future__ import annotations

from django.http import HttpRequest
from django.urls import reverse

from allauth.account.models import EmailAddress
from allauth.socialaccount.providers.base import ProviderAccount
from allauth.socialaccount.providers.oauth2.provider import OAuth2Provider
from allauth.socialaccount.providers.untappd.views import UntappdOAuth2Adapter


class UntappdAccount(ProviderAccount):
    def get_profile_url(self):
        return self.account.extra_data.get("untappd_url")

    def get_avatar_url(self):
        return self.account.extra_data.get("user_avatar")

    def get_user_data(self):
        return self.account.extra_data.get("response", {}).get("user", {})


class UntappdProvider(OAuth2Provider):
    id = "untappd"
    name = "Untappd"
    account_class = UntappdAccount
    oauth2_adapter_class = UntappdOAuth2Adapter

    def get_auth_params_from_request(self, request: HttpRequest, action):
        params = super().get_auth_params_from_request(request, action)
        # Untappd uses redirect_url instead of redirect_uri
        params["redirect_url"] = request.build_absolute_uri(
            reverse(f"{self.id}_callback")
        )
        return params

    def extract_uid(self, data):
        return str(data["response"]["user"]["uid"])

    def extract_common_fields(self, data):
        user = data["response"]["user"]
        return dict(
            username=user["user_name"],
            name=f"{user['first_name']} {user['last_name']}",
        )

    def extract_email_addresses(self, data):
        ret = [
            EmailAddress(
                email=data["response"]["user"]["settings"]["email_address"],
                verified=True,
                primary=True,
            )
        ]
        return ret


provider_classes = [UntappdProvider]



import posixpath
from urllib.parse import urlparse

from django.core.exceptions import ValidationError
from django.http import HttpRequest

from oauthlib.oauth2.rfc6749.errors import OAuth2Error

from allauth.idp.oidc.adapter import get_adapter


class InvalidTargetError(OAuth2Error):
    error = "invalid_target"
    description = "The requested resource is invalid, missing, unknown, or malformed."


def get_resources(request: HttpRequest) -> list[str]:
    params = request.GET if request.method == "GET" else request.POST
    resources = params.getlist("resource")
    validate_resources(resources)
    return resources


def validate_resources(resources: list[str]) -> None:
    for resource in resources:
        parsed = urlparse(resource)
        if not parsed.scheme or not parsed.netloc:
            raise ValidationError("Resource must be an absolute URI.")
        if parsed.fragment:
            raise ValidationError("Resource must not include a fragment component.")
        if parsed.path and posixpath.normpath(parsed.path) != parsed.path:
            raise ValidationError("Resource path is not normalized.")
        if "//" in parsed.path:
            raise ValidationError("Resource path is not normalized.")
    adapter = get_adapter()
    adapter.validate_resource_uris(uris=resources)


def is_resources_subset(requested: list[str] | None, granted: list[str] | None) -> bool:
    if not granted or not requested:
        return True
    for resource in requested:
        if not any(_is_contained_by(resource, g) for g in granted):
            return False
    return True


def _is_contained_by(resource: str, granted: str) -> bool:
    r = urlparse(resource)
    g = urlparse(granted)
    if r.scheme != g.scheme or r.netloc != g.netloc:
        return False
    g_path = g.path.rstrip("/") + "/"
    return r.path == g.path or r.path.startswith(g_path)
